import inspect
import re
from typing import List, Any
from abc import abstractmethod
from pydantic import BaseModel, validator, Field


class ParamDesc(BaseModel):
    name: str
    type: str
    required: bool


class ToolSchema(BaseModel):
    name: str
    description: str
    arguments: List[ParamDesc]

    
class BaseTool(BaseModel):
    name: str = Field("Name of the tool")
    description: str = Field("Description of what the tool does")

    @validator("name")
    def check_name(cls, v):
        # Satisfy OpenAI function calling requirements
        if not re.fullmatch("^[a-zA-Z0-9_-]+$'.", v):
            raise ValueError("Tool names must match pattern: '^[a-zA-Z0-9_-]+$'.")
        return v

    @abstractmethod
    def func(self, *args: Any, **kwargs: Any) -> Any:
        """Function of tool to be implemented by subclass"""
        raise NotImplementedError("Function of tool must be implemented by subclass")

    def get_schema(self):
        """Get schema of defined function for tool"""
        signature = inspect.signature(self.func)
        arguments = []
        for name, param in signature.parameters.items():
            required = True if param.default is inspect.Parameter.empty else False
            arguments.append(ParamDesc(name=name, type=param.annotation.__name__, required=required))
        return ToolSchema(name=self.name, description=self.description, arguments=arguments)


from typing import Generator, List, Optional, Union
from pydantic import BaseModel, Field, validator
from abc import abstractmethod

from compositeai.tools import BaseTool
from compositeai.drivers import BaseDriver


class AgentFinishTool(BaseTool):
    name: str = "agent_finish"
    description: str = "Return the final result once you believe you have completed the task at hand"

    def func(self, result: str) -> str:
        return result
    

class AgentResult(BaseModel):
    content: str = Field("Final result of agent execution")


class BaseAgent(BaseModel):
    driver: BaseDriver
    name: str = Field("Name of the AI agent")
    description: str = Field("Description of AI agent role")
    is_entry: Optional[bool] = Field(default=False, description="Setting to True makes this agent the entry point for the user")
    tools: Optional[List[BaseTool]] = Field(default=None)
    max_iterations: Optional[int] = Field(default=10, ge=0)


    @validator('tools', pre=True)
    def add_default_tools(cls, value):
        """Add default tools to tool list"""
        if not value:
            value = [AgentFinishTool()]
        else:
            value.append(AgentFinishTool())
        return value
    

    @abstractmethod
    def execute(self, task: str, input: Optional[str] = None, stream: bool = False) -> Union[Generator, AgentResult]:
        """Use driver LLM to generate a response (including function calling)"""
        raise NotImplementedError("Method must be implemented by a subclass")


from typing import List
from openai import OpenAI
from pydantic import validator, PrivateAttr
from dotenv import load_dotenv

from compositeai.drivers.base_driver import (
    BaseDriver,
    DriverUsage,
    DriverToolCall,
    DriverMessage,
    DriverResponse,
    DriverInput,
)
from compositeai.tools import BaseTool


class OpenAIDriver(BaseDriver):
    _client: OpenAI = PrivateAttr()

    
    def __init__(self, **data):
        super().__init__(**data)
        load_dotenv()
        self._client = OpenAI()


    @validator("model")
    def check_model(cls, v):
        _openai_supported_models = set([
            "gpt-4-turbo", 
            "gpt-4-turbo-2024-04-09", 
            "gpt-4-turbo-preview", 
            "gpt-4-0125-preview", 
            "gpt-4-1106-preview", 
            "gpt-4", 
            "gpt-4-0613", 
            "gpt-3.5-turbo", 
            "gpt-3.5-turbo-0125", 
            "gpt-3.5-turbo-1106", 
            "gpt-3.5-turbo-0613",
            "gpt-4o"
        ])
        if v not in _openai_supported_models:
            raise ValueError(f"Model must be one of {_openai_supported_models}.")
        return v
    
    
    def generate(
        self,
        input: DriverInput,
    ) -> DriverResponse:
        messages = input.messages
        messages = self._messages_driver_to_openai(messages)
        max_tokens = input.max_tokens
        temperature = input.temperature
        tools = input.tools
        if tools:
            tools = self._fc_schema_basetools_to_openai(tools)
        tool_choice = input.tool_choice
        if tool_choice:
            tool_choice = tool_choice.value

        response = self._client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_tokens,
            temperature=temperature,
            tools=tools,
            tool_choice=tool_choice,
        )

        content = response.choices[0].message.content
        tool_calls = response.choices[0].message.tool_calls 
        driver_tool_calls = None
        if tool_calls:
            driver_tool_calls = []
            for tool_call in tool_calls:
                driver_tool_call = DriverToolCall(id=tool_call.id, name=tool_call.function.name, args=tool_call.function.arguments)
                driver_tool_calls.append(driver_tool_call)
        usage = self._usage_openai_to_driver(response.usage)

        return DriverResponse(content=content, tool_calls=driver_tool_calls, usage=usage)


    def _messages_driver_to_openai(self, messages: List[DriverMessage]) -> List[object]:
        return [{"role": message.role.value, "content": message.content} for message in messages]
    

    # Helper function to convert BaseTool to OpenAI function calling schema
    def _fc_schema_basetools_to_openai(self, tools: List[BaseTool]) -> List[object]:
        openai_fcs = []
        for tool in tools:
            # Get schema of tool/function
            tool_schema = tool.get_schema()

            # OpenAI function calling tool format
            openai_fc = {
                "type": "function",
                "function": {
                    "name": tool_schema.name,
                    "description": tool_schema.description,
                    "parameters": {
                        "type": "object",
                        "properties": {
                        },
                        "required": []
                    },
                }
            }

            # Populate formatted arguments for tool
            for arg in tool_schema.arguments:
                openai_fc["function"]["parameters"]["properties"][arg.name] = {"type": self._type_conversion_openai(arg.type)}
                if arg.required:
                    openai_fc["function"]["parameters"]["required"].append(arg.name)
            
            openai_fcs.append(openai_fc)
        return openai_fcs
    

    # Helper function converting python types to OpenAI function calling type format
    def _type_conversion_openai(self, type: str) -> str:
        conversions = {
            "int": "integer",
            "str": "string",
            "float": "float",
            "bool": "boolean",
        }
        return conversions[type]
    

    def _usage_openai_to_driver(self, openai_usage_obj: object) -> DriverUsage:
        return DriverUsage(
            prompt_tokens=openai_usage_obj.prompt_tokens,
            completion_tokens=openai_usage_obj.completion_tokens,
            total_tokens=openai_usage_obj.total_tokens,
        )
    


from typing import Generator, List, Optional, Union
from pydantic import BaseModel, PrivateAttr
import textwrap
import json

from compositeai.agents.base_agent import AgentResult, AgentFinishTool, BaseAgent
from compositeai.drivers.base_driver import DriverInput, DriverToolChoice, DriverToolCall
from compositeai.tools import BaseTool


class RAISEAgent(BaseAgent):
    _scratchpad: List[str] = PrivateAttr(default=[])
    

    def execute(self, task: str, input: Optional[str] = None, stream: bool = False) -> Union[Generator, AgentResult]:
        # Add initial data to conversation history and agent scratchpad
        self._scratchpad.append(f"task: {task}")
        if input:
            self._scratchpad.append(input)

        # Loop through iterations
        for _ in range(self.max_iterations):
            for chunk in self._iterate():
                if stream:
                    yield chunk
                if isinstance(chunk, AgentResult):
                    return chunk

        # At this point, maximum number of iterations reached
        raise RuntimeError("Maximum number of iterations reached.")
    
    
    def refresh_memory(self):
        memory = f"""
            ~~~YOU ARE AN AI AGENT GIVEN THE FOLLOWING ROLE:
            {self.description}

            ~~~SO FAR YOUR SRATCHPAD CONTAINS THE FOLLOWING:
            {self._scratchpad}

            ~~~NOW PROCEED WITH NOVEL TASKS:
        """
        return textwrap.dedent(memory)
    

    def _iterate(self) -> Generator:
        memory = self.refresh_memory()
        plan = self._plan(
            system_prompt=memory,
            tools=self.tools,
        )
        self._scratchpad.append(
           f"ALL PREVIOUS PLANS ARE COMPLETE ***** NEW PLAN: {plan}")
        yield plan

        memory = self.refresh_memory()
        actions = self._action(
            system_prompt=memory,
        )
        self._scratchpad.append(f"action(s): Calling the following tools - {actions}")
        yield actions
        
        memory = self.refresh_memory()
        observation = self._observe(
            system_prompt=memory,
            actions=actions,
        )
        self._scratchpad.append(f"observation(s): {observation}")
        yield observation
            

    def _plan(
        self, 
        system_prompt: str, 
        tools: List[BaseTool],
    ) -> str:
        tools_json_desc = [tool.get_schema().json() for tool in tools]
        system_prompt = f"""
            {system_prompt}

            YOU CAN USE THE FOLLOWING TOOLS:
            {tools_json_desc}

            WRITE A BRIEF PLAN FOR WHAT YOU SHOULD DO AT THIS POINT IN TIME:
        """
        system_prompt = textwrap.dedent(system_prompt)
        messages = [{"role": "system", "content": system_prompt}]
        driver_input = DriverInput(
            messages=messages,
        )
        response = self.driver.generate(input=driver_input)
        content = response.content
        return content
    
    
    def _action(
        self, 
        system_prompt: str, 
    ) -> List[DriverToolCall]:
        system_prompt = f"""
            {system_prompt}

            USE THE GIVEN TOOLS TO EXECUTE YOUR PLAN.
        """
        system_prompt = textwrap.dedent(system_prompt)
        messages = [{"role": "system", "content": system_prompt}]
        driver_input = DriverInput(
            messages=messages,
            tools=self.tools,
            tool_choice=DriverToolChoice.REQUIRED
        )
        response = self.driver.generate(input=driver_input)
        tool_calls = response.tool_calls
        return tool_calls


    def _observe(
        self, 
        system_prompt: str,
        actions: List[DriverToolCall]
    ):
        data = []
        for tool_call in actions:
            # Get function call info
            function_name = tool_call.name
            function_args = json.loads(tool_call.args)

            # Iterate through provided tools to check if driver_response function call matches one
            no_match_flag = True
            for tool in self.tools:
                # If match, run tool function on arguments for result, and append to memory
                if tool.get_schema().name == function_name:
                    no_match_flag = False
                    function_result = str(tool.func(**function_args))
                    if function_name == "agent_finish":
                        return AgentResult(content=function_result)
                    data.append(function_result)
            
            # If driver_response function call matches none of the given tools
            if no_match_flag:
                raise Exception("Driver called function, function call does not match any of the provided tools.")
            
        # Once data has been obtained from the results of function calls, filter for useful insights as observations
        system_prompt = f"""
            {system_prompt}

            HERE ARE IS WHAT YOU OBSERVED FROM YOUR PREVIOUS ACTION(S):
            {data}

            EXTRACT THE MOST RELEVANT DATA FROM YOUR OBSERVATIONS:
        """
        system_prompt = textwrap.dedent(system_prompt)
        messages = [{"role": "system", "content": system_prompt}]
        driver_input = DriverInput(
            messages=messages
        )
        response = self.driver.generate(input=driver_input)
        content = response.content
        return content


from typing import Optional, List
from abc import abstractmethod
from pydantic import BaseModel, Field
from enum import Enum

from compositeai.tools import BaseTool


class DriverUsage(BaseModel):
    completion_tokens: int = Field("Completion tokens used for LLM response", ge=0)
    prompt_tokens: int = Field("Prompt tokens used to generate LLM response", ge=0)
    total_tokens: int = Field("Total tokens used to generate LLM response", ge=0)


class DriverToolCall(BaseModel):
    id: str = Field("ID of tool call")
    name: str = Field("Name of tool/function called")
    args: str = Field("JSON string of arguments")


class DriverResponse(BaseModel):
    content: Optional[str] = Field(default=None, description="Text generated by the LLM")
    tool_calls: Optional[List[DriverToolCall]] = Field(default=None, description="Tool calls from the LLM")
    usage: DriverUsage = Field("Usage data to generate LLM response")


class DriverToolChoice(Enum):
    NONE = 'none'
    AUTO = 'auto'
    REQUIRED = 'required'


class DriverRole(Enum):
    SYSTEM = 'system'
    USER = 'user'
    ASSISTANT = 'assistant'


class DriverMessage(BaseModel):
    role: DriverRole
    content: str


class DriverInput(BaseModel):
    messages: List[DriverMessage]
    max_tokens: Optional[int] = Field(default=None, ge=0)
    temperature: Optional[float] = Field(default=0.0, ge=0.0, le=2.0)
    tools: Optional[List[BaseTool]] = Field(default=None)
    tool_choice: Optional[DriverToolChoice] = Field(default=None)

    
class BaseDriver(BaseModel):
    model: str = Field("Name of the LLM model to use from specific driver")

    @abstractmethod
    def generate(
        self,
        input: DriverInput
    ) -> DriverResponse:
        """Use driver LLM to generate a response (including function calling)"""
        raise NotImplementedError("Method must be implemented by a subclass")
    


import requests
from bs4 import BeautifulSoup

from compositeai.tools import BaseTool


class WebScrapeTool(BaseTool):
    name: str = "scrape_website"
    description: str = "Scrape the content of a website given the URL as a string"

    def func(self, url: str) -> str:
        response = requests.get(url)
        if response.status_code == 200:
            soup = BeautifulSoup(response.content, "html.parser")
            text = soup.get_text()
            if len(text) > 16000:
                return "Requested content exceeds maximum length."
            return text
        else:
            return "Website scrape failed."


import requests
import json
import os
from typing import Any
from dotenv import load_dotenv

from compositeai.tools import BaseTool


class GoogleSerperApiTool(BaseTool):
    name: str = "google_search"
    description: str = "Retrieve Google search results using the Googler Serper API"


    def __init__(self, **data):
        super().__init__(**data)
        load_dotenv()
        self._SERP_API_KEY = os.getenv("SERP_API_KEY")


    def func(self, query: str) -> Any:
        url = "https://google.serper.dev/search"
        payload = json.dumps({
            "q": query
        })
        headers = {
            'X-API-KEY': self._SERP_API_KEY,
            'Content-Type': 'application/json'
        }
        response = requests.request("POST", url, headers=headers, data=payload)
        return response.json()["organic"]